from __future__ import division
# KL coefficient field representation
from dolfin import *
import numpy as np


class CoefField:
    """ artificial M-term KL
    """
    def __init__(self, mean, M, scale=1.0, decay=2.0):
        # type: (float, int, float, float) -> None
        """
         initialise field with given mean.
         field is log-uniform for expfield=True.
         length M of expansion can optionally be fixed in advance.
         usually it is determined by the length of y when evaluating realisations.
        :param mean: mean value of the field
        :param M: maximal number of terms in realisation
        :param decay: decay of the field
        :return: void
        """
        #                                           # create a Fenics expression of the affine field
        self.a = Expression('C + S*cos(A*pi*F1*x[0]) * cos(A*pi*F2*x[1])', A=2, C=mean, F1=0, F2=0, S=1, degree=5)
        self.mean = mean
        self.M = M
        self.scale = scale
        self.decay = decay

    def realisation(self, y, V):
        # type: (List[float], FunctionSpace) -> Function
        """
          evaluate realisation of random field subject to samples y and return interpolation onto FEM space V.
        :param y: list of samples of the RV in [-1, 1]
        :param V: FunctionSpace
        :return: Fenics Function as field realisation
        """

        def indexer(i):
            m1 = np.floor(i/2)
            m2 = np.ceil(i/2)
            return m1, m2
        assert self.M == len(y)                     # strong assumption but convenient
        a = self.a                                  # store affine field Expression

        
        
        a.C, a.S = self.mean, 0                     # get mean function as starting point
        x = interpolate(a, V).vector().array()      # interpolate constant mean on FunctionSpace
        a.C = 0                                     # set mean back to zero. From now on look only at amp_func
        #                                           # add up stochastic modes
        if self.M > 0:
            #                                       # get mean-length ratio
            for m, ym in enumerate(y):              # loop through sample items
                amp = (m+1) ** (-1 * self.decay)    # algebraic decay in front of the sin*sin
                a.F1, a.F2 = indexer(m+2)           # get running values in Expression
                a.S = self.scale                    # multiply a scaling parameter as well
                #                                   # add current Expression value
                x += amp * ym * interpolate(a, V).vector().array()
                
        f = Function(V)                             # create empty function on FunctionSpace
        #                                           # set function coefficients to realisation coefficients
        f.vector()[:] = x
        return f
    def get_am(self, m, V):
        """
          returns the m-th term of the affine coefficient representation
        :param m: term of the KL expansion, m=0 -> mean, m>0 -> KL term
        :param V: FunctionSpace
        :return: FEniCS Function
        """
        # type: (int, FunctionSpace) -> Function
        #one_fun = Function(V)
        #one_fun.vector()[:] = np.ones(V.dim())
        #return one_fun
        def indexer(i):
            m1 = np.floor(i/2)
            m2 = np.ceil(i/2)
            return m1, m2
        a = self.a                                  # store affine field Expression
        if m == 0:
            a.C, a.S = self.mean, 0                 # get mean function as starting point
            x = interpolate(a, V)                   # interpolate constant mean on FunctionSpace
            return x
        a.C = 0                                     # set mean back to zero. From now on look only at amp_func
        amp = (m+1) ** (-1 * self.decay)            # algebraic decay in front of the sin*sin
        a.F1, a.F2 = indexer(m+2)                   # get running values in Expression
        a.S = self.scale                            # multiply a scaling parameter as well
        #                                           # add current Expression value
        x = amp * interpolate(a, V)
        return x
